import unittest

class PaspTest(unittest.TestCase):
  def assertApproxEqual(self, X: list, Y: list, Z: list = None):
    self.assertEqual(len(X), len(Y))
    if Z is not None: self.assertEqual(len(Y), len(Z))
    for x, y in zip(X, Y): self.assertAlmostEqual(x, y)
    if Z is not None:
      for y, z in zip(Y, Z): self.assertAlmostEqual(y, z)




import sys, os

import pasp

if __name__ == "__main__":
  f = "earthquake" if len(sys.argv) < 2 else sys.argv[1]
  P = pasp.parse(f"examples/{f}.lp")
  print("=======\nProgram:\n=======")
  print(P)
  print("\n==============\nQuery results:\n==============")
  R = pasp.exact(P)
  # print("\n=================\nBC Query results:\n=================")
  # pasp.exact_bc(P)
  # print("\n=====================\nCredal Query results:\n=====================")
  # pasp.exact_sym(P)



import unittest
from .utils import PaspTest
import pasp

class TestExamples(PaspTest):
  def test_asia(self):
    P = pasp.parse("examples/asia.lp")
    R = pasp.exact(P)
    # ℙ(trip)
    self.assertApproxEqual(R[0], [0.01, 0.01])
    # ℙ(tuberculosis | trip)
    self.assertApproxEqual(R[1], [0.05, 0.05])
    # ℙ(cancer | smoking)
    self.assertApproxEqual(R[2], [0.1,  0.1 ])
    # ℙ(test | or)
    self.assertApproxEqual(R[3], [0.98, 0.98])
    # ℙ(smoking)
    self.assertApproxEqual(R[4], [0.5,  0.5 ])
    # ℙ(tuberculosis | not trip)
    self.assertApproxEqual(R[5], [0.01, 0.01])
    # ℙ(cancer | not smoking)
    self.assertApproxEqual(R[6], [0.01, 0.01])
    # ℙ(test | not or)
    self.assertApproxEqual(R[7], [0.05, 0.05])

  def test_earthquake(self):
    P = pasp.parse("examples/earthquake.lp")
    R = pasp.exact(P)
    # ℙ(alarm | burglary, earthquake)
    self.assertApproxEqual(R[0], [0.9,  0.9])
    # ℙ(alarm | not burglary, earthquake)
    self.assertApproxEqual(R[1], [0.1,  0.1])
    # ℙ(alarm | burglary, not earthquake)
    self.assertApproxEqual(R[2], [0.8,  0.8])
    # ℙ(alarm | not burglary, not earthquake)
    self.assertApproxEqual(R[3], [0.0,  0.0])

  def test_game(self):
    P = pasp.parse("examples/game.lp")
    R = pasp.exact(P)
    # ℙ(wins(b))
    self.assertApproxEqual(R[0], [0.7,  1.0])
    # ℙ(wins(c))
    self.assertApproxEqual(R[1], [0.3,  0.3])

  def test_insomnia(self):
    P = pasp.parse("examples/insomnia.lp")
    R = pasp.exact(P)
    # ℙ(insomnia)
    self.assertApproxEqual(R[0], [0.3,  0.3])
    # ℙ(work)
    self.assertApproxEqual(R[1], [0.3,  1.0])
    # ℙ(sleep)
    self.assertApproxEqual(R[2], [0.0,  0.7])
    # ℙ(not sleep)
    self.assertApproxEqual(R[3], [0.3,  1.0])
    # ℙ(not work)
    self.assertApproxEqual(R[4], [0.0,  0.7])

  def test_prisoners(self):
    P = pasp.parse("examples/prisoners.lp")
    R = pasp.exact(P)
    α = 19/40
    # ℙ(e1 | u)
    print(R[0], [1.0/(1+2*(((1-α)/α)**2)), 1.0/(1+2*((α/(1-α))**2))])
    self.assertApproxEqual(R[0], [1.0/(1+2*(((1-α)/α)**2)), 1.0/(1+2*((α/(1-α))**2))])
    # ℙ(e1 | not b, u)
    self.assertApproxEqual(R[1], [1.0/(1+((1-α)/α)**2), 1.0/(1+((α/(1-α))**2))])
    # ℙ(g | e1, u)
    self.assertApproxEqual(R[2], [0.0, 1.0])
    # ℙ(d)
    self.assertApproxEqual(R[3], [0.0, 1.0])
    # ℙ(e1 | g, u)
    self.assertApproxEqual(R[4], [0.0, 1.0/(1+(α/(1-α))**2)])
    # ℙ(e1 | ga, u)
    self.assertApproxEqual(R[5], [1.0/(1+(((1-α)/α)**2)/α), 1.0/(1+((α/(1-α))**2)/(1-α))])

  def test_simple(self):
    P = pasp.parse("examples/simple.lp")
    R = pasp.exact(P)
    # ℙ(s(a))
    self.assertApproxEqual(R[0], [0.20, 0.20])
    # ℙ(s(b))
    self.assertApproxEqual(R[1], [0.30, 0.30])
    # ℙ(v)
    self.assertApproxEqual(R[2], [0.048, 0.048])

  def test_simpler(self):
    P = pasp.parse("examples/simpler.lp")
    R = pasp.exact(P)
    # ℙ(r)
    self.assertApproxEqual(R[0], [0.5, 0.5])
    # ℙ(v)
    self.assertApproxEqual(R[1], [0.25, 0.25])

  def test_smokers(self):
    P = pasp.parse("examples/smokers.lp")
    R = pasp.exact(P)
    # ℙ(smokes(a))
    self.assertApproxEqual(R[0], [0.06, 0.06])
    # ℙ(smokes(b))
    self.assertApproxEqual(R[1], [0.2, 0.2])

class TestLStable(PaspTest):
  def test_barber(self):
    P = pasp.parse("examples/barber.lp", semantics = "lstable")
    R = pasp.exact(P)
    # ℙ(shaves(b, a))
    self.assertApproxEqual(R[0], [1.0, 1.0])
    # ℙ(not shaves(b, b))
    self.assertApproxEqual(R[1], [0.5, 0.5])
    # ℙ(undef shaves(b, b))
    self.assertApproxEqual(R[2], [0.5, 0.5])

class TestPlog(PaspTest):
  def test_asia(self):
    P = pasp.parse("examples/asia.lp")
    R = pasp.exact(P, psemantics = "plog")
    # ℙ(trip)
    self.assertApproxEqual(R[0], [0.01, 0.01])
    # ℙ(tuberculosis | trip)
    self.assertApproxEqual(R[1], [0.05, 0.05])
    # ℙ(cancer | smoking)
    self.assertApproxEqual(R[2], [0.1,  0.1 ])
    # ℙ(test | or)
    self.assertApproxEqual(R[3], [0.98, 0.98])
    # ℙ(smoking)
    self.assertApproxEqual(R[4], [0.5,  0.5 ])
    # ℙ(tuberculosis | not trip)
    self.assertApproxEqual(R[5], [0.01, 0.01])
    # ℙ(cancer | not smoking)
    self.assertApproxEqual(R[6], [0.01, 0.01])
    # ℙ(test | not or)
    self.assertApproxEqual(R[7], [0.05, 0.05])

  def test_earthquake(self):
    P = pasp.parse("examples/earthquake.lp")
    R = pasp.exact(P, psemantics = "plog")
    # ℙ(alarm | burglary, earthquake)
    self.assertApproxEqual(R[0], [0.9,  0.9])
    # ℙ(alarm | not burglary, earthquake)
    self.assertApproxEqual(R[1], [0.1,  0.1])
    # ℙ(alarm | burglary, not earthquake)
    self.assertApproxEqual(R[2], [0.8,  0.8])
    # ℙ(alarm | not burglary, not earthquake)
    self.assertApproxEqual(R[3], [0.0,  0.0])

  def test_simple(self):
    P = pasp.parse("examples/simple.lp")
    R = pasp.exact(P, psemantics = "plog")
    # ℙ(s(a))
    self.assertApproxEqual(R[0], [0.20, 0.20])
    # ℙ(s(b))
    self.assertApproxEqual(R[1], [0.30, 0.30])
    # ℙ(v)
    self.assertApproxEqual(R[2], [0.048, 0.048])

  def test_simpler(self):
    P = pasp.parse("examples/simpler.lp")
    R = pasp.exact(P, psemantics = "plog")
    # ℙ(r)
    self.assertApproxEqual(R[0], [0.5, 0.5])
    # ℙ(v)
    self.assertApproxEqual(R[1], [0.25, 0.25])

  def test_smokers(self):
    P = pasp.parse("examples/smokers.lp")
    R = pasp.exact(P, psemantics = "plog")
    # ℙ(smokes(a))
    self.assertApproxEqual(R[0], [0.06, 0.06])
    # ℙ(smokes(b))
    self.assertApproxEqual(R[1], [0.2, 0.2])

  def test_game(self):
    P = pasp.parse("examples/game.lp")
    R = pasp.exact(P, psemantics = "plog")
    # ℙ(wins(b))
    self.assertApproxEqual(R[0], [1.7/2, 1.7/2])
    # ℙ(wins(c))
    self.assertApproxEqual(R[1], [0.3,  0.3])

  def test_insomnia(self):
    P = pasp.parse("examples/insomnia.lp")
    R = pasp.exact(P, psemantics = "plog")
    # ℙ(insomnia)
    self.assertApproxEqual(R[0], [0.3,  0.3])
    # ℙ(work)
    self.assertApproxEqual(R[1], [0.3+(1.0-0.3)/2,  0.3+(1.0-0.3)/2])
    # ℙ(sleep)
    self.assertApproxEqual(R[2], [0.7/2,  0.7/2])
    # ℙ(not sleep)
    self.assertApproxEqual(R[3], [0.3+(1.0-0.3)/2,  0.3+(1.0-0.3)/2])
    # ℙ(not work)
    self.assertApproxEqual(R[4], [0.7/2,  0.7/2])

if __name__ == "__main__":
  unittest.main()



from .program import ProbFact

def pr(phi: list[ProbFact], theta: iter) -> float:
  """
  Returns the probability of the total choice θ asserted through the probabilistic facts ϕ.
  """
  n = len(phi)
  p = 1.0
  for i in range(n):
    q, t = float(phi[i].p), theta[i]
    p *= t*q + (not t)*(1-q) # ϕ[i].p if θ[i] is true, otherwise 1-ϕ[i].p.
  return p



import pathlib, enum
import lark, lark.reconstruct
from .program import ProbFact, Query, ProbRule, Program, CredalFact, unique_fact, Semantics

def is_fact(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a fact."
  return isinstance(x, lark.Tree) and x.data == "fact"
def is_pfact(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a probabilistic fact."
  return isinstance(x, lark.Tree) and x.data == "pfact"
def is_rule(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a rule."
  return isinstance(x, lark.Tree) and x.data == "rule"
def is_prule(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a probabilistic rule."
  return isinstance(x, lark.Tree) and x.data == "prule"
def is_atom(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a(n) (grounded) atom."
  return isinstance(x, lark.Tree) and (x.data == "atom" or x.data == "gratom")
def is_pred(x: lark.Tree) -> bool:
  "Returns whether a node in the AST is a (grounded) predicate."
  return isinstance(x, lark.Tree) and (x.data == "pred" or x.data == "grpred")

def atom_sign(x: lark.Tree) -> bool:
  "Returns whether a(n) (grounded) atom is negative (`False`), positive (`True`), or not an atom (`None`)."
  return x.children[0].type != "NEG" if is_atom(x) else None
def pred_sign(x: lark.Tree) -> bool:
  "Returns whether a (grounded) predicate is negative (`False`), positive (`True`) or not a predicate (`None`)."
  return x.children[0].type != "NEG" if is_pred(x) else None

def is_prob(x: lark.Tree) -> iter:
  "Returns whether a node (in this case a fact or rule) is probabilistic."
  return isinstance(x, lark.Tree) and len(x.children) > 0 and isinstance(x.children[0], lark.Token) and x.children[0].type == "PROB"

def facts(x: lark.Tree) -> iter:
  "Returns an iterator containing all facts in the AST."
  return x.find_pred(is_fact)
def pfacts(x: lark.Tree) -> iter:
  "Returns an iterator containing all probabilistic facts in the AST."
  return x.find_pred(is_pfact)
def rules(x: lark.Tree) -> iter:
  "Returns an iterator containing all rules in the AST."
  return x.find_pred(is_rule)
def prules(x: lark.Tree) -> iter:
  "Returns an iterator containing all probabilistic rules in the AST."
  return x.find_pred(is_prule)
def probs(x: lark.Tree) -> iter:
  "Returns an iterator containing all probabilistic facts and rules in the AST."
  return x.find_pred(lambda x: is_prule(x) or is_pfact(x))

def expand_interval(x: lark.Tree) -> tuple[int, int]:
  assert isinstance(x, lark.Tree) and x.data == "interval", "Given AST node is not an Interval."
  return int(x.children[0].value), int(x.children[1].value)

def tree_contains(x: lark.Tree, f) -> bool:
  "Performs a depth-first search, returning (and stopping the search) node `y` when `f(y) == True`."
  V = set()
  def visit(x: lark.Tree | lark.Token) -> bool:
    V.add(x)
    if f(x): return True
    if isinstance(x, lark.Tree):
      for c in x.children:
        if (c not in V) and visit(c): return True
    return False
  return visit(x)

def is_var(x: lark.Token): return isinstance(x, lark.Token) and x.type == "VAR"

def is_nonground(x: lark.Tree) -> bool:
  "Returns whether node `x` is not grounded, i.e. whether any node in the subtree of `x` is a variable."
  try: iter(x)
  except TypeError: return tree_contains(x, is_var)
  else: return any(tree_contains(x, is_var))
def is_ground(x: lark.Tree) -> bool:
  "Returns whether node `x` is grounded, i.e. whether no node in the subtree of `x` is a variable."
  return not is_nonground(x)

def read(*files: str, G: lark.Lark = None, from_str: bool = False) -> lark.Tree:
  "Read all `files` and parse them with grammar `G`, returning a single `lark.Tree`."
  if G is None:
    try:
      with open(pathlib.Path(__file__).resolve().parent.joinpath("grammar.lark"), "r") as f:
        G = lark.Lark(f, start = "plp")
    except Exception as ex:
      raise ex
  T = None
  if from_str: T = G.parse("\n".join(files)); return T
  for fname in files:
    try:
      # For now, dump files entirely into memory (this isn't too much of a problem, since PLPs are
      # usually small). In the future, consider streaming batches of text instead for large files.
      with open(fname, "r") as f:
        if T is None: T = G.parse(f.read())
        else:
          U = G.parse(f.read())
          T.children.extend(u for u in U.children if u not in T.children)
    except Exception as ex:
      raise ex
  assert T is not None, "No file read."
  return T

def getnths(X: iter, i: int) -> iter: return (x[i] for x in X)

class Command(enum.Enum):
  FACT = 0
  PROB_FACT = 1
  RULE = 2
  PROB_RULE = 3
  QUERY = 4
  CRED_FACT = 5
  CONSTRAINT = 6
  CONSTDEF = 7

class PLPTransformer(lark.Transformer):
  def __init__(self, _):
    super().__init__()
    self.semantics = Semantics.STABLE

  # Terminals.
  def UND(self, a: list[lark.Token]) -> tuple[str, bool]: return a, True
  def CONST(self, a: list[lark.Token]) -> tuple[str, bool]: return a, True
  def NEG(self, a: list[lark.Token]) -> tuple[str, bool]: return a, True
  def VAR(self, a: list[lark.Token]) -> tuple[str, bool]: return a, False
  def ID(self, a: list[lark.Token]) -> tuple[str, bool]: return a, True
  def OP(self, a: list[lark.Token]) -> tuple[str, bool]: return a, True

  # Atoms.
  def atom(self, a: list[lark.Tree]) -> tuple[str, bool]:
    return " ".join(getnths(a, 0)), all(getnths(a, 1)), True
  def gratom(self, a: list[lark.Tree]) -> tuple[str, bool]: return self.atom(a)[0], True

  # Intervals.
  def interval(self, i: list[lark.Token]) -> tuple[str, bool]:
    return "..".join(getnths(i, 0)), all(getnths(i, 1))

  # Predicates.
  def pred(self, p: list[tuple[str, bool]]) -> tuple[str, bool]:
    sign = p[0][0].type == "NEG"
    if sign:
      name, args = p[1], p[2:]
      return f"not {name}({', '.join(getnths(args, 0))})"
    name, args = p[0], p[1:]
    return f"{name[0]}({', '.join(getnths(args, 0))})", all(getnths(args, 1)), True
  def grpred(self, p: list[tuple[str, bool]]) -> tuple[str, bool]: return self.pred(p)[0], True

  # Binary operations.
  def bop(self, b: list[lark.Tree]) -> str:
    return " ".join(getnths(b, 0)), all(getnths(b, 1)), False

  # Facts.
  def fact(self, f: list[lark.Tree]) -> tuple[Command, str]:
    return Command.FACT, "".join(getnths(f, 0)) + "."
  def pfact(self, f: list[lark.Tree]) -> tuple[Command, ProbFact]:
    return Command.PROB_FACT, ProbFact(f[0], f[1][0])
  def cfact(self, f: list[lark.Tree]) -> tuple[Command, CredalFact]:
    return Command.CRED_FACT, CredalFact(str(f[0]), str(f[1]), f[2][0])

  # Heads.
  def head(self, h: list[str]) -> str:
    return ", ".join(getnths(h, 0)), all(getnths(h, 1)), [x for x in getnths(h, 0)]
  def ohead(self, h: list[str]) -> str:
    if len(h) == 1: return str(h[0][0]), True, h[0][0], ()
    u = h[1:]
    return f"{h[0][0]}({', '.join(getnths(u, 0))})", all(getnths(u, 1)), h[0][0], list(getnths(u, 0))
  # Bodies.
  def body(self, b: list[str]) -> str:
    return ", ".join(getnths(b, 0)), all(getnths(b, 1)), list(x for x, t in zip(getnths(b, 0), getnths(b, 2)) if t), list(getnths(b, 0))

  # Rules.
  def rule(self, r: list[str]) -> tuple[Command, str]:
    return Command.RULE, f"{r[0][0]} :- {r[1][0]}."
  def prule(self, r: list[str]) -> tuple[Command, str, ProbFact]:
    o = f"{r[1][0]} :- {r[2][0]}"
    prop = r[1][1] and r[2][1]
    if prop: return Command.PROB_RULE, ProbRule(r[0], o, is_prop = True)
    h, b = r[1][3], r[2][2]
    # Invariant: len(b) > 0, otherwise the rule is unsafe.
    h_s = ", ".join(h) + ", " if len(h) > 0 else ""
    b_s = ", ".join(map(lambda x: f"0, {x[4:]}" if x[:4] == "not " else f"1, {x}", b))
    u = f"{r[1][2]}(@unify(\"{r[0]}\", {r[1][2]}, {len(h)}, {2*len(b)}, {h_s}{b_s})) :- {r[2][0]}."
    return Command.PROB_RULE, ProbRule(r[0], o, is_prop = False, unify = u)

  # Constraint.
  def constraint(self, b: list[str]) -> tuple[Command, str]:
    return Command.CONSTRAINT, f":- {b[0][0]}."

  # Interpretations.
  def interp(self, i: list[str]) -> list[str]: return list(getnths(i, 0))

  # Queries.
  def query(self, q: list[list[str]]) -> tuple[str, Query]:
    return Command.QUERY, Query(q[0], q[1] if len(q) > 1 else [], semantics = self.semantics)

  def constdef(self, t: list) -> tuple[Command, str]:
    return Command.CONSTDEF, f"#const {t[0][0]} = {t[1][0]}."

  # Probabilistic Logic Program.
  def plp(self, C: list[tuple]) -> Program:
    # Logic Program.
    P  = []
    # Probabilistic Facts.
    PF = []
    # Probabilistic Rules.
    PR = []
    # Queries.
    Q  = []
    # Credal Facts.
    CF = []
    for t, *c in C:
      if t == Command.FACT: P.extend(c)
      elif t == Command.PROB_FACT: PF.extend(c)
      elif t == Command.RULE: P.extend(c)
      elif t == Command.PROB_RULE:
        for r in c:
          PR.append(r)
          if r.is_prop:
            P.append(r.prop_f)
            PF.append(r.prop_pf)
      elif t == Command.QUERY: Q.extend(c)
      elif t == Command.CRED_FACT: CF.extend(c)
      elif t == Command.CONSTRAINT: P.extend(c)
      else: P.extend(c)
    return Program("\n".join(P), PF, PR, Q, CF, semantics = self.semantics)

class PartialTransformer(PLPTransformer):
  def __init__(self, sem: str):
    super().__init__(sem)
    self.PT = set()
    self.semantics = Semantics.LSTABLE if sem == "lstable" else Semantics.PARTIAL
    self.o_tree = None

  @staticmethod
  def has_binop(x: str): return ("=" in x) or ("<" in x) or (">" in x)

  def fact(self, f) -> tuple[Command, str]:
    u = "".join(getnths(f, 0))
    self.PT.add(u)
    return Command.FACT, u + "."

  def pfact(self, f) -> tuple[Command, ProbFact]:
    self.PT.add(f[1][0])
    return Command.PROB_FACT, ProbFact(f[0], f[1][0])

  def cfact(self, f) -> tuple[Command, CredalFact]:
    self.PT.add(f[2][0])
    return Command.CRED_FACT, CredalFact(str(f[0]), str(f[1]), f[2][0])

  def rule(self, r) -> tuple[Command, str, str, str]:
    b1 = ", ".join(map(lambda x: f"not _{x[4:]}" if x[:4] == "not " else x, r[1][3]))
    b2 = ", ".join(map(lambda x: x if (x[:4] == "not ") or PartialTransformer.has_binop(x) else f"_{x}", r[1][3]))
    h1, h2 = ", ".join(r[0][2]), ", ".join(map(lambda x: f"_{x}", r[0][2]))
    for h in r[0][2]: self.PT.add(h)
    # for x in r[1][3]:
      # if not PartialTransformer.has_binop(x): self.PT.add(x[4:] if x[:4] == "not " else x)
    return Command.RULE, f"{h1} :- {b1}.", f"{h2} :- {b2}."

  def prule(self, r) -> tuple[Command, str, str, str]:
    tr_negs = lambda x: f"not _{x[4:]}" if x[:4] == "not " else x
    tr_pos  = lambda x: x if (x[:4] == "not ") or PartialTransformer.has_binop(x) else f"_{x}"
    b1 = ", ".join(map(tr_negs, r[2][3]))
    b2 = ", ".join(map(tr_pos, r[2][3]))
    h = r[1][0]
    o1, o2 = f"{h} :- {b1}", f"_{h} :- {b2}"
    self.PT.add(h)
    # for x in r[2][3]:
      # if not PartialTransformer.has_binop(x): self.PT.add(x[4:] if x[:4] == "not " else x)
    prop = r[1][1] and r[2][1]
    uid = unique_fact()
    if prop: return Command.PROB_RULE, ProbRule(r[0], o1, ufact = uid), ProbRule(r[0], o2, ufact = uid)
    h_a, b = r[1][3], r[2][2]
    # Invariant: len(b) > 0, otherwise the rule is unsafe.
    h_s = ", ".join(h_a) + ", " if len(h_a) > 0 else ""
    b1_s = ", ".join(map(lambda x: f"0, _{x[4:]}" if x[:4] == "not " else f"1, {x}", b))
    # Let the grounder deal with the _f rule.
    u1 = f"{r[1][2]}(@unify(\"{r[0]}\", {r[1][2]}, {len(h_a)}, {2*len(b)}, {h_s}{b1_s})) :- {b1}."
    return Command.PROB_RULE, ProbRule(r[0], o1, is_prop = False, unify = u1)

  def plp(self, C: list[tuple]) -> Program:
    # Logic Program.
    P  = []
    # Probabilistic Facts.
    PF = []
    # Probabilistic Rules.
    PR = []
    # Queries.
    Q  = []
    # Credal Facts.
    CF = []
    for t, *c in C:
      if t == Command.FACT: P.extend(c)
      elif t == Command.PROB_FACT: PF.extend(c)
      elif t == Command.RULE: P.extend(c)
      elif t == Command.PROB_RULE:
        if len(c) == 1:
          PR.append(c[0])
          continue
        r1, r2 = c
        PR.extend((r1, r2))
        if r1.is_prop:
          P.append(r1.prop_f)
          PF.append(r1.prop_pf)
        if r2.is_prop:
          P.append(r2.prop_f)
      elif t == Command.QUERY: Q.extend(c)
      elif t == Command.CRED_FACT: CF.extend(c)
      elif t == Command.CONSTRAINT: P.extend(c)
      else: P.extend(c)
    P.extend(f"_{x} :- {x}." for x in self.PT)
    return Program("\n".join(P), PF, PR, Q, CF, semantics = self.semantics, \
                   stable_p = self.stable_p)

  def transform(self, tree):
    self.o_tree = tree
    self.stable_p = PLPTransformer(self.semantics).transform(tree)
    return super().transform(tree)

def parse(*files: str, G: lark.Lark = None, from_str: bool = False, semantics: str = "stable") -> Program:
  """Either parses `streams` as blocks of text containing the PLP when `from_str = True`, or
  interprets `streams` as filenames to be read and parsed into a `Program`."""
  if semantics not in parse.trans_map:
    raise ValueError("semantics not supported (must either be 'stable', 'partial' or 'lstable')!")
  return parse.trans_map[semantics](semantics).transform(read(*files, G = G, from_str = from_str))
parse.trans_map = {}
parse.trans_map["stable"] = PLPTransformer
parse.trans_map["lstable"] = PartialTransformer
parse.trans_map["partial"] = PartialTransformer




# cython: c_string_type=unicode, c_string_encoding=utf8

import enum

import clingo
from clingo.symbol import Function

def unique_fact(i: int = None) -> str:
  """
  Creates a new unique fact for probabilistic rules. To do this, we update a counter `unique_fact.i`
  in a way equivalent to C's `static` variables.
  """
  if i is None:
    unique_fact.i += 1
    return f"__unique_id_{unique_fact.i}"
  return f"__unique_id_{i}"
unique_fact.i = 0

class ProbFact:
  """
  A Probabilistic Fact (PF) is a (Logic Program) fact which is "chosen" with some probability.
  """

  def __init__(self, p: str, f: str):
    "Constructs a PF out of a probability `p` and fact `f`."
    self.p = float(p)
    self.f = f
    # Construct a clingo.symbol.Function from this fact.
    self.cl_f = clingo.parse_term(f)

  def __str__(self) -> str: return f"{self.p}::{self.f}"
  def __repr__(self) -> str: return self.__str__()

class ProbRule:
  """
  A Probabilistic Rule (PR) is a (Logic Program) rule that (when propositional) may be chosen with
  some probability `p`. A non-propositional PR must be grounded first.
  """

  def __init__(self, p: str, f: str, is_prop: bool = True, unify: str = None, ufact: str = None, \
               partial: str = None):
    self.p = p
    self.f = f
    self.is_prop = is_prop
    self.unify = unify
    self.prop_pf = ProbFact(p, unique_fact() if ufact is None else ufact)
    self.prop_f = f"{f}, {self.prop_pf.f}."
    #self.partial = f"{partial}, {self.prop_pf.f}."

  def __str__(self) -> str: return f"{self.p}::{self.f}"
  def __repr__(self) -> str: return self.__str__()

class CredalFact:
  """
  A Credal Fact (CF) consists of a fact `f` attached to a probability interval `[l, u]`, where `l ∈
  [0, 1]` is the lowest probability `f` may attain and `u ≥ l ∈ [0, 1]` is the highest.
  """

  def __init__(self, l: float, u: float, f: str):
    self.l, self.u = float(l), float(u)
    self.f = f
    self.cl_f = clingo.parse_term(f)

  def __getitem__(self, i: bool) -> float: return self.l if False else self.u
  def __str__(self) -> str: return f"[{self.l}, {self.u}]::{self.f}"
  def __repr__(self) -> str: return self.__str__()

def _str_query_assignment(f: Function, t: bool) -> str:
  """
  String formats a query tuple `(f, t)`, where `f` is an atom and `t` is whether it should appear
  in the program or not.
  """
  return str(f) if t == Query.TERM_POS else ("not " + str(f) if t == Query.TERM_NEG else "undef " + str(f))

class Semantics(enum.IntEnum):
  STABLE = 0
  PARTIAL = 1
  LSTABLE = 2

class Query:
  """
  A query is a meta-command within a PLP to signal the solver to produce and output a probabilistic
  query. A query follows a modified PASOCS [1] syntax, that is, the query (not necessarily in this
  order)

  ```
  #query(q1; ...; qk; not p1; ...; not pm | e1; ...; en; not v1; ...; not vt)
  ```

  of ground atoms `q1, ..., qk`, `p1, ..., pm`, `e1, ..., en`, `v1, ..., vt` is equivalent to asking
  the probability

  ```
  P({q1, ..., qk} = true, {p1, ..., pm} = false | {e1, ..., en} = true, {v1, ..., vt} = false).
  ```

  See concrete examples in the `/examples` folder.

  [1] - PASOCS: A Parallel Approximate Solver for Probabilistic Logic Programs under the Credal
  Semantics. Tuckey et al, 2021. URL: https://arxiv.org/abs/2105.10908.
  """

  TERM_NEG = 0
  TERM_POS = 1
  TERM_UND = 2

  def __init__(self, Q: iter, E: iter = [], semantics: Semantics = Semantics.STABLE):
    """
    Constructs a query from query (`Q`) and evidence (`E`) assignments.

    We use the notation `iter` as a type hinting to mean `Q` and `E` are iterables.
    """
    self.Q = [Query.parse_term(q, semantics) for q in Q]
    self.E = [Query.parse_term(e, semantics) for e in E]

  @staticmethod
  def parse_term(u: str, s: Semantics):
    if u.startswith("not "): t, n = u[4:], Query.TERM_NEG
    elif u.startswith("undef "): t, n = u[6:], Query.TERM_UND
    else: t, n = u, Query.TERM_POS
    return clingo.parse_term(t), n, None if s == Semantics.STABLE else clingo.parse_term(f"_{t}")

  def __str__(self) -> str:
    qs = f"ℙ({', '.join(_str_query_assignment(q, t) for q, t, _ in self.Q)}"
    if len(self.E) != 0: return qs + f" | {', '.join(_str_query_assignment(e, t) for e, t, _ in self.E)})"
    return qs + ")"
  def __repr__(self) -> str: return self.__str__()

class Program:
  """
  A Probabilistic Logic Program (PLP) usually configures a triple `<P,PF,CF>`, where `P` is a logic
  program, `PF` are probabilistic facts and `CF` are credal facts. We extend a PLP into a tuple
  `<P,PF,CF,Q>`, where `Q` are the queries to be asked from `P`, `PF` and `CF`.

  We accept ProbLog's syntactic sugar for probabilistic rules,

  ```
  p::h(X) :- b1(X), b2(X), ..., bn(X).
  ```

  meaning that if `b1(X), b2(X), ..., bn(X)` is true, `h(X)` is added with probability `p`. This is
  equivalent to

  ```
  p::a.
  h(X) :- b1(X), b2(X), ..., bn(X), a.
  ```

  where `a` is a unique probabilistic fact added with probability `p`.
  """

  def __init__(self, P: str, PF: list[ProbFact], PR: list[ProbRule], Q: list[Query], \
               CF: list[CredalFact], semantics: Semantics = Semantics.STABLE, stable_p = None):
    """
    Constructs a PLP out of a logic program `P`, probabilistic facts `PF`, credal facts `CF` and
    queries `Q`.
    """
    self.P = P
    self.PF = PF
    self.PR = PR
    self.Q = Q
    self.CF = CF

    self.gr_P = None
    self.gr_PF = None
    self.gr_pr = None

    self.semantics = semantics
    self.stable = stable_p

  def __str__(self) -> str:
    return f"<Logic Program:\n{self.P},\nProbabilistic Facts:\n{self.PF},\nCredal Facts:\n{self.CF}\nProbabilistic Rules:\n{self.PR},\nQueries\n{self.Q}>"
  def __repr__(self) -> str: return self.__str__()



import os
from setuptools import setup, Extension, find_packages, Command

class TestCommand(Command):
  description = "Runs unit and functional tests for PASP."
  user_options = []

  def initialize_options(self):
    self.cwd = None

  def finalize_options(self):
    self.cwd = os.getcwd()

  def run(self):
    assert os.getcwd() == self.cwd, f"Must be in package root: {self.cwd}"
    os.system("python setup.py build_ext --inplace && python -m unittest tests/examples.py -b")

coptimize = Extension("coptimize",
                      depends = ["pasp/coptimize.h"],
                      sources = ["pasp/coptimize.c"])
cutils    = Extension("cutils",
                      libraries = ["clingo"],
                      depends = ["pasp/cutils.h"],
                      sources = ["pasp/cutils.c"])
cprogram  = Extension("cprogram",
                      libraries = ["clingo"],
                      depends = ["pasp/cutils.c", "pasp/cprogram.h"],
                      sources = ["pasp/cprogram.c"])
cexact    = Extension("cexact",
                      libraries = ["m", "clingo", "pthread"],
                      depends = ["pasp/cprogram.c", "pasp/coptimize.c", "pasp/cinf.c"],
                      sources = ["pasp/cexact.c", "thpool/thpool.c", "pasp/cinf.c"],
                      extra_compile_args = ["-Wno-unused-function"],
                      define_macros = [("NUM_PROCS", str(os.cpu_count())), ("_GNU_SOURCE", None)])
carray    = Extension("carray",
                      depends = ["pasp/carray.h"],
                      sources = ["pasp/carray.c"])
cground   = Extension("cground",
                      libraries = ["clingo"],
                      depends = ["pasp/cutils.c", "pasp/cprogram.c", "pasp/cground.h"],
                      sources = ["pasp/cground.c"])

setup(
  packages = find_packages(where = ".", include = ["pasp*"]),
  include_package_data = True,
  ext_modules = [coptimize, cutils, cprogram, cexact, carray, cground],
  cmdclass = {"test": TestCommand},
)



import sys
import os
import contextlib
import array
import timeit

import clingo

class DummyStream:
  def write(self, x): pass

_global_dummy_stream = DummyStream()

@contextlib.contextmanager
def suppress_out():
  _stdout = sys.stdout
  sys.stdout = _global_dummy_stream
  yield
  sys.stdout = _stdout

@contextlib.contextmanager
def suppress_ext_out():
  with open(os.devnull, "w") as devnull:
    _stdout = sys.stdout
    sys.stdout = devnull
    try:
      yield
    finally:
      sys.stdout = _stdout

@contextlib.contextmanager
def suppress_err():
  _stderr = sys.stderr
  sys.stderr = _global_dummy_stream
  yield
  sys.stderr = _stderr

@contextlib.contextmanager
def suppress_ext_err():
  with open(os.devnull, "w") as devnull:
    _stderr = sys.stderr
    sys.stderr = devnull
    try:
      yield
    finally:
      sys.stderr = _stderr

@contextlib.contextmanager
def timer_block():
  s = timeit.default_timer()
  try:
    yield
  finally:
    print("Time elapsed: ", timeit.default_timer() - s)

def start_timer() -> float: start_timer.s = timeit.default_timer(); return start_timer.s
def end_timer() -> float: return timeit.default_timer() - start_timer.s

def undef_atom_ignore(x, y):
  if x == clingo.MessageCode.AtomUndefined: return
  print(y, file = sys.stderr)

def new_array(n: int, t: type = 'b', v: int = 0): return array.array(t, [v])*n
def new_list(n: int, v = None): return [v for _ in range(n)]
